@@ -28,15 +28,9 @@ from typing import (
     Any,
     Callable,
     ContextManager,
-    DefaultDict,
-    Dict,
     Iterator,
-    List,
     Mapping,
     MutableSet,
-    Optional,
-    Set,
-    Tuple,
 )
 from weakref import WeakSet
 
@@ -54,6 +48,7 @@ from synapse.logging.context import (
 from synapse.logging.opentracing import start_active_span
 from synapse.metrics import SERVER_NAME_LABEL, Histogram, LaterGauge
 from synapse.util.clock import Clock
+from synapse.util.duration import Duration
 
 if typing.TYPE_CHECKING:
     from contextlib import _GeneratorContextManager
@@ -104,7 +99,7 @@ _rate_limiter_instances_lock = threading.Lock()
 
 def _get_counts_from_rate_limiter_instance(
     count_func: Callable[["FederationRateLimiter"], int],
-) -> Mapping[Tuple[str, ...], int]:
+) -> Mapping[tuple[str, ...], int]:
     """Returns a count of something (slept/rejected hosts) by (metrics_name)"""
     # Cast to a list to prevent it changing while the Prometheus
     # thread is collecting metrics
@@ -114,7 +109,7 @@ def _get_counts_from_rate_limiter_instance(
     # Map from (metrics_name,) -> int, the number of something like slept hosts
     # or rejected hosts. The key type is Tuple[str], but we leave the length
     # unspecified for compatability with LaterGauge's annotations.
-    counts: Dict[Tuple[str, ...], int] = {}
+    counts: dict[tuple[str, ...], int] = {}
     for rate_limiter_instance in rate_limiter_instances:
         # Only track metrics if they provided a `metrics_name` to
         # differentiate this instance of the rate limiter.
@@ -169,7 +164,7 @@ class FederationRateLimiter:
         our_server_name: str,
         clock: Clock,
         config: FederationRatelimitSettings,
-        metrics_name: Optional[str] = None,
+        metrics_name: str | None = None,
     ):
         """
         Args:
@@ -191,7 +186,7 @@ class FederationRateLimiter:
                 metrics_name=metrics_name,
             )
 
-        self.ratelimiters: DefaultDict[str, "_PerHostRatelimiter"] = (
+        self.ratelimiters: collections.defaultdict[str, "_PerHostRatelimiter"] = (
             collections.defaultdict(new_limiter)
         )
 
@@ -222,7 +217,7 @@ class _PerHostRatelimiter:
         our_server_name: str,
         clock: Clock,
         config: FederationRatelimitSettings,
-        metrics_name: Optional[str] = None,
+        metrics_name: str | None = None,
     ):
         """
         Args:
@@ -244,7 +239,7 @@ class _PerHostRatelimiter:
         self.concurrent_requests = config.concurrent
 
         # request_id objects for requests which have been slept
-        self.sleeping_requests: Set[object] = set()
+        self.sleeping_requests: set[object] = set()
 
         # map from request_id object to Deferred for requests which are ready
         # for processing but have been queued
@@ -253,11 +248,11 @@ class _PerHostRatelimiter:
         ] = collections.OrderedDict()
 
         # request id objects for requests which are in progress
-        self.current_processing: Set[object] = set()
+        self.current_processing: set[object] = set()
 
         # times at which we have recently (within the last window_size ms)
         # received requests.
-        self.request_times: List[int] = []
+        self.request_times: list[int] = []
 
     @contextlib.contextmanager
     def ratelimit(self, host: str) -> "Iterator[defer.Deferred[None]]":
@@ -359,7 +354,9 @@ class _PerHostRatelimiter:
                     rate_limiter_name=self.metrics_name,
                     **{SERVER_NAME_LABEL: self.our_server_name},
                 ).inc()
-            ret_defer = run_in_background(self.clock.sleep, self.sleep_sec)
+            ret_defer = run_in_background(
+                self.clock.sleep, Duration(seconds=self.sleep_sec)
+            )
 
             self.sleeping_requests.add(request_id)
 
@@ -420,6 +417,6 @@ class _PerHostRatelimiter:
                 pass
 
         self.clock.call_later(
-            0.0,
+            Duration(seconds=0),
             start_next_request,
         )
